private import
	std.string,
	dyayaml;

alias void delegate(char[] s) PutS;
alias YamlHolder function(YamlLangContext context) YamlLangBuiltInFunction;
alias YamlHolder delegate(YamlLangContext context) YamlLangBuiltInDelegate;

class YamlLangError : Error 
{
	this(char[] message){ super(message); }
}

YamlHolder example_of_false, example_of_true;

static this()
{
	example_of_false = new YamlScalar(``);
	example_of_true = new YamlScalar(`1`);
}

abstract class YamlLangContext
{
	this(YamlLangContext _parent){
		parent = _parent;
	}

	YamlLangContext find_method(YamlLangContext _parent, char[] name)
	{
		if(parent !== null){
			return parent.find_method(_parent, name);
		}else{
			return null;
		}
	}
	
	abstract YamlHolder execute();
	
	YamlHolder get_variable(char[] name)
	{
		YamlHolder result;
		if(name in variables){
			result = variables[name];
		}else if(parent !== null){
			result = parent.get_variable(name);
		}
		if(result === null){
			throw new YamlLangError(`variable "` ~ name ~ `" not found.`);
		}
		return result;
	}
	
	void puts(char[] s)
	{
		assert(parent !== null);
		parent.puts(s);
	}
	
	YamlLangContext parent;
	YamlHolder[char[]] variables;
}

class YamlLangBuiltInFunctionRunner : YamlLangContext
{
	this(YamlLangContext _parent, YamlLangBuiltInFunction _method){
		super(_parent);
		method = _method;
	}
	
	override YamlHolder execute()
	{
		YamlHolder result = method(this);
		assert(result !== null);
		return result;
	}
	
	YamlLangBuiltInFunction method;
}

class YamlLangBuiltInDelegateRunner : YamlLangContext
{
	this(YamlLangContext _parent, YamlLangBuiltInDelegate _method){
		super(_parent);
		method = _method;
	}
	
	override YamlHolder execute()
	{
		YamlHolder result = method(this);
		assert(result !== null);
		return result;
	}
	
	YamlLangBuiltInDelegate method;
}

class YamlLangMethodRunner : YamlLangContext
{
	this(YamlLangContext _parent, YamlHolder _base){
		super(_parent);
		base = _base;
	}
	
	override YamlLangContext find_method(YamlLangContext _parent, char[] name)
	{
		YamlLangContext result;
		if(base.is_hash()){
			if(name in base.hash()){
				result = new YamlLangMethodRunner(_parent, base[name]);
			}
		}
		if(result === null && parent !== null){
			result = parent.find_method(_parent, name);
		}
		return result;
	}
	
	YamlHolder execute(YamlHolder entry)
	{
		YamlHolder result;
		if(entry.is_hash()){
			foreach(char[] key, Yaml_Element value; entry.hash()){
				//printf("%.*s,", key);
				if(key.length > 0 && key[0] == '@'){
					char[] funcname = key[1..key.length];
					//printf("@%.*s ", funcname);
					YamlLangContext method = find_method(this, funcname);
					if(method === null){
						YamlHolder closure;
						try{
							closure = get_variable(funcname);
						}catch(YamlLangError e){
							throw new YamlLangError(`method or ` ~ e.msg);
						}
						method = new YamlLangMethodRunner(this, closure);
					}
					if(value.is_hash()){
						foreach(char[] argname, Yaml_Element expression; value.hash()){
							if(expression.is_list()){
								method.variables[argname] = expression;
							}else{
								method.variables[argname] = execute(expression);
							}
						}
					}
					result = method.execute();
					assert(result !== null);
					break;
				}
			}
			//printf("\n");
			assert(result !== null);
		}else if(entry.is_list()){
			//printf("[");
			int length = entry.list().length;
			assert(length > 0);
			for(int i = 0; i < length; ++i){
				result = execute(entry[i]);
				assert(result !== null);
			}
			//printf("]");
		}else{
			result = entry;
			char[] s = result.string_value();
			if(s.length > 0 && s[0] == '@'){
				result = get_variable(s[1..s.length]);
			}
		}
		if(result === null){
			throw new YamlLangError(`result has not assigned.`);
		}
		return result;
	}
	
	override YamlHolder execute()
	{
		return execute(base);
	}
	
	YamlHolder base;
}

class YamlLangRunner : YamlLangMethodRunner
{
	this(YamlLangRunner _parent, YamlHolder _base)
	{
		super(_parent, _base);
	}
	
	override YamlLangContext find_method(YamlLangContext _parent, char[] name)
	{
		YamlLangContext result = super.find_method(_parent, name);
		if(result === null){
			if(name in built_in_functions){
				result = new YamlLangBuiltInFunctionRunner(_parent, built_in_functions[name]);
			}else if(name in built_in_delegates){
				result = new YamlLangBuiltInDelegateRunner(_parent, built_in_delegates[name]);
			}
		}
		return result;
	}
	
	YamlHolder execute(char[] message)
	{
		assert(message.length > 0);
		assert(message[0] == '@');
		char[] funcname = message[1..message.length];
		YamlLangContext method = find_method(this, funcname);
		if(method === null){
			throw new YamlLangError(`"` ~ funcname ~ `" not found.`);
		}
		return method.execute();
	}
	
	void ready_print()
	{
		built_in_functions[`print`] = function YamlHolder(YamlLangContext context){
			context.puts(context.get_variable(`msg`).string_value());
			return example_of_false;
		};
	}
	
	void ready_if()
	{
		built_in_functions[`if`] = function YamlHolder(YamlLangContext context){
			YamlHolder result;
			YamlHolder cond = context.get_variable(`cond`);
			//printf("cond:%.*s\n", cond.value());
			if(cond.bit_value()){
				YamlHolder then = context.get_variable(`then`);
				result = (new YamlLangMethodRunner(context, then)).execute();
			}else{
				result = example_of_false;
			}
			return result;
		};
	}
	
	void ready_nequal()
	{
		built_in_functions[`nequal`] = function YamlHolder(YamlLangContext context){
			YamlHolder lhs = context.get_variable(`lhs`);
			YamlHolder rhs = context.get_variable(`rhs`);
			assert(lhs.is_scalar);
			assert(rhs.is_scalar);
			//printf("lhs:%.*s\n", lhs.value());
			//printf("rhs:%.*s\n", rhs.value());
			return (lhs.string_value() != rhs.string_value()) ?
				example_of_true : example_of_false;
		};
	}
	
	void ready_add()
	{
		built_in_functions[`add`] = function YamlHolder(YamlLangContext context){
			YamlHolder lhs = context.get_variable(`lhs`);
			YamlHolder rhs = context.get_variable(`rhs`);
			assert(lhs.is_scalar);
			assert(rhs.is_scalar);
			return new YamlScalar(std.string.toString(lhs.integer_value() + rhs.integer_value()));
		};
	}
	
	void set_default_built_in_methods(PutS puts)
	{
		puts_field = puts;
		ready_print();
		ready_if();
		ready_nequal();
		ready_add();
	}
	
	override void puts(char[] s)
	{
		puts_field(s);
	}
	
	PutS puts_field;
	YamlLangBuiltInFunction[char[]] built_in_functions;
	YamlLangBuiltInDelegate[char[]] built_in_delegates;
}
